import gc
import os
import pickle
import threading
import time
import logging

from timewise.parent_sample_base import ParentSampleBase

logger = logging.getLogger(__name__)


class BigParentSampleBase(ParentSampleBase):
    """
    This should not be used. It was a bad idea. The better way would be to implement a
    ParentSample class that splits the sample file in separate files and then maybe use Dask or similar.
    """

    def __init__(self, base_name, keep_file_in_memory=30*60):
        """
        See doc of ParentSampleBase

        :param base_name: base name for storage directories
        :type base_name: str
        :param keep_file_in_memory: time in seconds to keep the parent sample file in the memory, after that gets written to a cache file on disk
        :type keep_file_in_memory: float
        """
        super().__init__(base_name=base_name)

        self._keep_df_in_memory = keep_file_in_memory
        self._time_when_df_was_used_last = time.time()
        self._df = None
        self._cache_file = os.path.join(self.cache_dir, "cache.pkl")
        self._lock_cache_file = False

        self._clean_thread = threading.Thread(target=self._periodically_drop_df_to_disk, daemon=True, name='ParentSampleCleanThread').start()
        self._stop_thread = False

    def _wait_for_unlock_cache_file(self):
        if self._lock_cache_file:
            logger.debug('cache file locked, waiting')
            while self._lock_cache_file:
                pass

        logger.debug('cache file unlocked')

    @property
    def df(self):
        self._time_when_df_was_used_last = time.time()

        if isinstance(self._df, type(None)):

            if os.path.isfile(self._cache_file):
                logger.debug(f'loading from {self._cache_file}')
                self._wait_for_unlock_cache_file()
                self._lock_cache_file = True

                with open(self._cache_file, "rb") as f:
                    self._df = pickle.load(f)

                self._lock_cache_file = False

            else:
                logger.debug(f'No file {self._cache_file}')

        return self._df

    @df.setter
    def df(self, value):
        self._time_when_df_was_used_last = time.time()
        self._df = value

    def _periodically_drop_df_to_disk(self):
        logger.debug(f'starting cleaning thread for parent sample')

        while True:
            if self._stop_thread:
                break

            if (
                    ((time.time() - self._time_when_df_was_used_last) > self._keep_df_in_memory)
                    and not isinstance(self._df, type(None))
            ):
                logger.debug(f'writing DataFrame to {self._cache_file}')
                self._wait_for_unlock_cache_file()
                self._lock_cache_file = True

                with open(self._cache_file, "wb") as f:
                    pickle.dump(self._df, f)

                self._lock_cache_file = False
                self._df = None
                gc.collect()

            time.sleep(self._keep_df_in_memory / 2)

        logger.debug('stopped clean thread')

    def __del__(self):
        if hasattr(self, "_cache_file") and os.path.isfile(self._cache_file):
            logger.debug(f'removing {self._cache_file}')
            os.remove(self._cache_file)

        if hasattr(self, "clean_thread"):
            logger.debug(f'stopping clean thread')
            self._stop_thread = True
            self._clean_thread.join()


import requests
import os
import getpass
import logging
import pandas as pd
import matplotlib.pyplot as plt
import pyvo as vo
import backoff
import numpy
from astropy.table import Table
from PIL import Image
from io import BytesIO


from timewise.general import cache_dir, backoff_hndlr


logger = logging.getLogger(__name__)
mirong_url = 'http://staff.ustc.edu.cn/~jnac/data_public/wisevar.txt'
local_copy = os.path.join(cache_dir, 'mirong_sample.csv')


def get_mirong_sample():

    if not os.path.isfile(local_copy):

        logger.info(f'getting MIRONG sample from {mirong_url}')
        r = requests.get(mirong_url)
        lll = list()
        for l in r.text.split('\n')[1:]:
            illl = list()
            for ll in l.split(' '):
                if ll and '#' not in ll:
                    illl.append(ll)
            lll.append(illl)

        mirong_sample = pd.DataFrame(lll[1:-1], columns=lll[0])
        mirong_sample['ra'] = mirong_sample['RA']
        mirong_sample['dec'] = mirong_sample['DEC']
        logger.debug(f'saving to {local_copy}')

        mirong_sample.to_csv(local_copy, index=False)
        logger.info(f'found {len(mirong_sample)} objects in MIRONG Sample')
        mirong_sample.drop(columns=['ra', 'dec'], inplace=True)
        mirong_sample.to_csv(local_copy, index=False)

    else:
        logger.debug(f'loading {local_copy}')
        mirong_sample = pd.read_csv(local_copy)

    return mirong_sample


###########################################################################################################
#            START SDSS UTILS                       #
#####################################################


def get_sdss_credentials():
    if not os.environ.get('SDSS_USERID'):
        os.environ['SDSS_USERID'] = input('Enter SDSS user ID:')
    if not os.environ.get('SDSS_USERPW'):
        os.environ['SDSS_USERPW'] = getpass.getpass('Enter SDSS password:')
    return os.environ['SDSS_USERID'], os.environ['SDSS_USERPW']


def login_to_sciserver():
    try:
        from SciServer import Authentication
    except ModuleNotFoundError:
        raise ModuleNotFoundError("Please install SciServer (https://github.com/sciserver/SciScript-Python) "
                                  "if you want to see SDSS cutouts!")

    uid, pw = get_sdss_credentials()
    logger.debug(f"logging in to SciServer with username {uid}")
    Authentication.login(uid, pw)


@backoff.on_exception(
        backoff.expo,
        requests.RequestException,
        max_tries=50,
        on_backoff=backoff_hndlr
    )
def get_cutout(*args, **kwargs):
    login_to_sciserver()
    from SciServer import SkyServer
    return SkyServer.getJpegImgCutout(*args, **kwargs)


def plot_sdss_cutout(ra, dec, arcsec=20, arcsec_per_px=0.1, interactive=False, fn=None, title=None, save=False, ax=False,
                height=2.5):

    ang_px = int(arcsec / arcsec_per_px)
    ang_deg = arcsec / 3600

    if not ax:
        fig, ax = plt.subplots(figsize=(height, height))
    else:
        fig = plt.gcf()

    try:
        im = get_cutout(ra, dec, scale=arcsec_per_px, height=ang_px, width=ang_px)
        ax.imshow(im, origin='upper',
                  extent=([ra + ang_deg / 2, ra - ang_deg / 2,
                           dec - ang_deg / 2, dec + ang_deg / 2]),
                  cmap='gray')
        ax.scatter(ra, dec, marker='x', color='red')

    except Exception as e:
        if "outside the SDSS footprint" in str(e):
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            x = sum(xlim) / 2
            y = sum(ylim) / 2
            ax.annotate("Outside SDSS Footprint", (x, y), color='red', ha='center', va='center', fontsize=20)
        else:
            raise Exception(e)

    if title:
        ax.set_title(title)

    if save:
        logger.debug(f"saving under {fn}")
        fig.savefig(fn)

    if interactive:
        return fig, ax

    plt.close()


#####################################################
#            END SDSS UTILS                         #
###########################################################################################################

###########################################################################################################
#            START PANSTARRS UTILS                  #
#####################################################


def getimages(ra, dec, filters="grizy"):
    """Query ps1filenames.py service to get a list of images

    ra, dec = position in degrees
    size = image size in pixels (0.25 arcsec/pixel)
    filters = string with filters to include
    Returns a table with the results
    """

    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = f"{service}?ra={ra}&dec={dec}&filters={filters}"
    table = Table.read(url, format='ascii')
    return table


def geturl(ra, dec, size=240, output_size=None, filters="grizy", format="jpg", color=False):
    """Get URL for images in the table

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filters = string with filters to include
    format = data format (options are "jpg", "png" or "fits")
    color = if True, creates a color image (only for jpg or png format).
            Default is return a list of URLs for single-filter grayscale images.
    Returns a string with the URL
    """

    if color and format == "fits":
        raise ValueError("color images are available only for jpg or png formats")
    if format not in ("jpg", "png", "fits"):
        raise ValueError("format must be one of jpg, png, fits")
    table = getimages(ra, dec, filters=filters)
    url = (f"https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
           f"ra={ra}&dec={dec}&size={size}&format={format}")
    if output_size:
        url = url + "&output_size={}".format(output_size)
    # sort filters from red to blue
    flist = ["yzirg".find(x) for x in table['filter']]
    table = table[numpy.argsort(flist)]
    if color:
        if len(table) > 3:
            # pick 3 filters
            table = table[[0, len(table) // 2, len(table) - 1]]
        for i, param in enumerate(["red", "green", "blue"]):
            url = url + "&{}={}".format(param, table['filename'][i])
    else:
        urlbase = url + "&red="
        url = []
        for filename in table['filename']:
            url.append(urlbase + filename)
    return url


def getcolorim(ra, dec, size=240, output_size=None, filters="grizy", format="jpg"):
    """Get color image at a sky position

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filters = string with filters to include
    format = data format (options are "jpg", "png")
    Returns the image
    """

    if format not in ("jpg", "png"):
        raise ValueError("format must be jpg or png")
    url = geturl(ra, dec, size=size, filters=filters, output_size=output_size, format=format, color=True)
    r = requests.get(url)
    im = Image.open(BytesIO(r.content))
    return im


def getgrayim(ra, dec, size=240, output_size=None, filter="g", format="jpg"):
    """Get grayscale image at a sky position

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filter = string with filter to extract (one of grizy)
    format = data format (options are "jpg", "png")
    Returns the image
    """

    if format not in ("jpg", "png"):
        raise ValueError("format must be jpg or png")
    if filter not in list("grizy"):
        raise ValueError("filter must be one of grizy")
    url = geturl(ra, dec, size=size, filters=filter, output_size=output_size, format=format)
    r = requests.get(url[0])
    im = Image.open(BytesIO(r.content))
    return im


def plot_panstarrs_cutout(
        ra,
        dec,
        arcsec,
        interactive=False,
        fn=None,
        title=None,
        save=False,
        ax=False,
        plot_color_image=False,
        height=2.5
):
    arcsec_per_px = 0.25
    ang_px = int(arcsec / arcsec_per_px)

    imshow_kwargs = {
        'origin': 'upper',
        "extent": ([arcsec / 2, -arcsec / 2, -arcsec / 2, arcsec / 2])
    }
    scatter_args = [0, 0]
    scatter_kwargs = {'marker': 'x', 'color': 'red'}

    if not plot_color_image:
        filters = 'grizy'
        if not ax:
            fig, axss = plt.subplots(2, len(filters), sharex='all', sharey='all',
                                     gridspec_kw={'wspace': 0, 'hspace': 0, 'height_ratios': [1, 8]},
                                     figsize=(height * 5, height))
        else:
            fig = plt.gcf()
            axss = ax

        for j, fil in enumerate(list(filters)):
            im = getgrayim(ra, dec, size=ang_px, filter=fil)
            axs = axss[1]
            axs[j].imshow(im, cmap='gray', **imshow_kwargs)

            axs[j].scatter(*scatter_args, **scatter_kwargs)
            axs[j].set_title(fil)
            axss[0][j].axis('off')

    else:
        logger.debug('plotting color image')
        if not ax:
            fig, axss = plt.subplots(figsize=(height, height))
        else:
            fig = plt.gcf()
            axss = ax

        im = getcolorim(ra, dec, size=ang_px)
        axss.imshow(im, **imshow_kwargs)
        axss.scatter(*scatter_args, **scatter_kwargs)

    _this_title = title if title else f"{ra}_{dec}"
    si = "-" if dec > 0 else "+"
    ylabel = f"Dec {si} {abs(dec):.2f} [arcsec]"
    xlabel = f"RA - {ra:.2f} [arcsec]"
    try:
        axss.set_title(_this_title)
        axss.set_xlabel(xlabel)
        axss.set_ylabel(ylabel)
        axss.grid(ls=":", alpha=0.5)
    except AttributeError:  # in this case axss is an array
        fig.supylabel(ylabel)
        fig.supxlabel(xlabel)
        fig.suptitle(_this_title)
        for a in axss.flatten():
            a.grid(ls=":", alpha=0.5)

    if save:
        logger.info(f'saving under {fn}')
        fig.savefig(fn)

    if interactive:
        return fig, axss

    plt.close()


#####################################################
#            END PANSTARRS UTILS                    #
###########################################################################################################

###########################################################################################################
#            START EXCESS VARIANCE UTILS            #
#####################################################

# calculate excess variance as done in section 7.3.4 of (Boller et al)
# https://www.aanda.org/articles/aa/full_html/2016/04/aa25648-15/aa25648-15.html#S26


def calc_Expectation(a):
     
    # variable prb is for probability
    # of each element which is same for
    # each element
    n = len(a)
    prb = 1 / n
    # calculating expectation overall
    sum = 0
    for i in range(0, n):
        sum += (a[i] * prb)
    # returning expectation as sum
    return float(sum)


def get_excess_variance(y, y_err, mu):
    import numpy as np
    N = len(y)
    sum_variance = 0
    for i, (X, sig) in enumerate(zip(y, y_err)):
        sum_variance += np.power(X-mu,2) - np.power(sig,2)
       
    excess_variance = (sum_variance)/(N*mu**2)
    
    # calculate the uncertainty
    F_var = np.sqrt(np.abs(excess_variance))/mu
    std_exp = calc_Expectation(y_err**2)
    term1 = np.sqrt(2/N)*std_exp/(np.power(mu,2))
    term2 = np.sqrt((std_exp*2*F_var)/(N*mu))
    
    uncertainty = np.sqrt(term1**2 + term2**2)
    return excess_variance, uncertainty


#####################################################
#              END EXCESS VARIANCE UTILS            #
###########################################################################################################


###########################################################################################################
#            START CUSTOM TAP Service                 #
#######################################################


class StableAsyncTAPJob(vo.dal.AsyncTAPJob):
    """
    Implements backoff for call of phase which otherwise breaks the code if there are connection issues
    """

    @property
    @backoff.on_exception(
        backoff.expo,
        (vo.dal.DALServiceError, AttributeError),
        max_tries=50,
        on_backoff=backoff_hndlr
    )
    def phase(self):
        return super(StableAsyncTAPJob, self).phase


class StableTAPService(vo.dal.TAPService):
    """
    Implements the StableAsyncTAPJob for job submission
    """

    def submit_job(
            self,
            query,
            language="ADQL",
            maxrec=None,
            uploads=None,
            **keywords
    ):
        return StableAsyncTAPJob.create(
            self.baseurl,
            query,
            language,
            maxrec,
            uploads,
            self._session,
            **keywords
        )


#######################################################
#            END CUSTOM TAP Service                   #
###########################################################################################################



import os
import pandas as pd
import numpy as np
import logging
from scipy import stats
import matplotlib.pyplot as plt

from timewise.wise_data_base import WISEDataBase
from timewise.utils import get_excess_variance

logger = logging.getLogger(__name__)


class WiseDataByVisit(WISEDataBase):
    """
    WISEData class to bin lightcurve by visit
    """

    mean_key = '_mean'
    median_key = '_median'
    rms_key = '_rms'
    upper_limit_key = '_ul'
    Npoints_key = '_Npoints'
    zeropoint_key_ext = '_zeropoint'

    flux_error_factor = {
        "W1": 1.6,
        "W2": 1.2
    }

    def __init__(
            self,
            base_name,
            parent_sample_class,
            min_sep_arcsec,
            n_chunks,
            clean_outliers_when_binning=True,
            multiply_flux_error=True
    ):
        # TODO: add doc
        super().__init__(base_name, parent_sample_class, min_sep_arcsec, n_chunks)
        self.clean_outliers_when_binning = clean_outliers_when_binning
        self.multiply_flux_error = multiply_flux_error

    def calculate_epochs(self, f, e, visit_mask, counts, remove_outliers, outlier_mask=None):
        """
        Calculates the binned epochs of a lightcurve.

        :param f: the fluxes
        :type f: np.array
        :param e: the flux errors
        :type e: np.array
        :param visit_mask: the visit mask
        :type visit_mask: np.array
        :param counts: the counts
        :type counts: np.array
        :param remove_outliers: whether to remove outliers
        :type remove_outliers: bool
        :param outlier_mask: the outlier mask
        :type outlier_mask: np.array
        :return: the epoch
        :rtype: float
        """
        u_lims = pd.isna(e)
        nan_mask = pd.isna(f)

        # ---------------------   remove outliers in the bins   ---------------------- #

        # if we do not want to clean outliers just set the threshold to infinity
        outlier_thresh = np.inf if not remove_outliers else 20

        # set up empty masks
        outlier_mask = np.array([False] * len(f)) if outlier_mask is None else outlier_mask
        median = np.nan
        u = np.nan
        use_mask = None
        n_points = counts

        # set up dummy values for number of remaining outliers
        n_remaining_outlier = np.inf

        # ---------------------   flag upper limits   ---------------------- #
        bin_n_ulims = np.bincount(visit_mask, weights=u_lims, minlength=len(counts))
        bin_ulim_bool = (counts - bin_n_ulims) == 0
        use_mask_ul = ~u_lims | (u_lims & bin_ulim_bool[visit_mask])

        n_loops = 0

        # recalculate uncertainty and median as long as no outliers left
        while n_remaining_outlier > 0:

            # make a mask of values to use
            use_mask = ~outlier_mask & use_mask_ul & ~nan_mask
            n_points = np.bincount(visit_mask, weights=use_mask)
            zero_points_mask = n_points == 0

            # -------------------------   calculate median   ------------------------- #
            median = np.zeros_like(counts, dtype=float)
            visits_at_least_one_point = np.unique(visit_mask[~zero_points_mask[visit_mask]])
            visits_zero_points = np.unique(visit_mask[zero_points_mask[visit_mask]])
            median[visits_at_least_one_point] = np.array([
                np.median(f[(visit_mask == i) & use_mask]) for i in visits_at_least_one_point
            ])
            median[visits_zero_points] = np.nan

            # median is NaN for visits with 0 detections, (i.e. detections in one band and not the other)
            # if median is NaN for other visits raise Error
            if np.any(np.isnan(median[n_points > 0])):
                nan_indices = np.where(np.isnan(median))[0]
                msg = ""
                for inan_index in nan_indices:
                    nanf = f[visit_mask == inan_index]
                    msg += f"median is nan for {inan_index}th bin\n{nanf}\n\n"
                raise ValueError(msg)

            # ---------------------   calculate uncertainty   ---------------------- #
            mean_deviation = np.bincount(
                visit_mask[use_mask],
                weights=(f[use_mask] - median[visit_mask[use_mask]]) ** 2,
                minlength=len(counts)
            )
            one_points_mask = n_points <= 1
            # calculate standard deviation
            std = np.zeros_like(counts, dtype=float)
            std[~one_points_mask] = (
                    np.sqrt(mean_deviation[~one_points_mask])
                    / (n_points[~one_points_mask] - 1)
                    * stats.t.interval(0.68, df=n_points[~one_points_mask] - 1)[1]
                    # for visits with small number of detections we have to correct according to the t distribution
            )
            std[one_points_mask] = -np.inf

            # calculate the propagated errors of the single exposure measurements
            single_exp_measurement_errors = np.sqrt(np.bincount(
                visit_mask[use_mask],
                weights=e[use_mask] ** 2,
                minlength=len(counts)
            ))
            e_meas = np.zeros_like(std, dtype=float)
            e_meas[~zero_points_mask] = single_exp_measurement_errors[n_points > 0] / n_points[n_points > 0]
            e_meas[zero_points_mask] = np.nan
            # take the maximum value of the measured single exposure errors and the standard deviation
            u = np.maximum(std, e_meas)

            # calculate 90% confidence interval
            u70 = np.zeros_like(counts, dtype=float)
            u70[one_points_mask] = 0
            visits_at_least_two_point = np.unique(visit_mask[~one_points_mask[visit_mask]])
            u70[visits_at_least_two_point] = np.array([
                np.quantile(abs(f[(visit_mask == i) & use_mask] - median[i]), .7, method="interpolated_inverted_cdf")
                for i in visits_at_least_two_point
            ])

            # ---------------------   remove outliers in the bins   ---------------------- #
            remaining_outliers = (abs(median[visit_mask] - f) > outlier_thresh * u70[visit_mask]) & ~outlier_mask
            outlier_mask |= remaining_outliers
            n_remaining_outlier = sum(remaining_outliers) if remove_outliers else 0
            # setting remaining_outliers to 0 will exit the while loop

            n_loops += 1

            if n_loops > 20:
                raise Exception(f"{n_loops}!")

        return median, u, bin_ulim_bool, outlier_mask, use_mask, n_points

    @staticmethod
    def get_visit_map(lightcurve):
        """
        Create a map datapoint to visit

        :param lightcurve: the unbinned lightcurve
        :type lightcurve: pd.DataFrame
        :returns: visit map
        :rtype: np.ndarray
        """
        # -------------------------   find epoch intervals   -------------------------- #
        sorted_mjds = np.sort(lightcurve.mjd)
        epoch_bounds_mask = (sorted_mjds[1:] - sorted_mjds[:-1]) > 100
        epoch_bins = np.array(
            [lightcurve.mjd.min() * 0.99] +  # this makes sure that the first datapoint gets selected
            list(((sorted_mjds[1:] + sorted_mjds[:-1]) / 2)[epoch_bounds_mask]) +  # finding the middle between
                                                                                   # two visits
            [lightcurve.mjd.max() * 1.01]    # this just makes sure that the last datapoint gets selected as well
        )

        visit_mask = np.digitize(lightcurve.mjd, epoch_bins) - 1
        return visit_mask

    def bin_lightcurve(self, lightcurve):
        """
        Combine the data by visits of the satellite of one region in the sky.
        The visits typically consist of some tens of observations. The individual visits are separated by about
        six months.
        The mean flux for one visit is calculated by the weighted mean of the data.
        The error on that mean is calculated by the root-mean-squared and corrected by the t-value.
        Outliers per visit are identified if they are more than 100 times the rms away from the mean. These outliers
        are removed from the calculation of the mean and the error if self.clean_outliers_when_binning is True.

        :param lightcurve: the unbinned lightcurve
        :type lightcurve: pandas.DataFrame
        :return: the binned lightcurve
        :rtype: pandas.DataFrame
        """

        # -------------------------   create visit mask   -------------------------- #
        visit_map = self.get_visit_map(lightcurve)
        counts = np.bincount(visit_map)

        binned_data = dict()

        # -------------------------   calculate mean mjd   -------------------------- #
        binned_data["mean_mjd"] = np.bincount(visit_map, weights=lightcurve.mjd) / counts

        # -------------------------   loop through bands   -------------------------- #
        for b in self.bands:
            # loop through magnitude and flux and save the respective datapoints

            outlier_masks = dict()
            use_masks = dict()
            bin_ulim_bools = dict()

            for lum_ext in [self.flux_key_ext, self.mag_key_ext]:
                f = lightcurve[f"{b}{lum_ext}"]
                e = lightcurve[f"{b}{lum_ext}{self.error_key_ext}"]

                # we will flag outliers based on the flux only
                remove_outliers = lum_ext == self.flux_key_ext and self.clean_outliers_when_binning
                outlier_mask = outlier_masks.get(self.flux_key_ext, None)

                mean, u, bin_ulim_bool, outlier_mask, use_mask, n_points = self.calculate_epochs(
                    f, e, visit_map,
                    counts,
                    remove_outliers=remove_outliers,
                    outlier_mask=outlier_mask
                )
                n_outliers = np.sum(outlier_mask)

                if (n_outliers > 0):
                    logger.info(f"removed {n_outliers} outliers")

                binned_data[f'{b}{self.mean_key}{lum_ext}'] = mean
                binned_data[f'{b}{lum_ext}{self.rms_key}'] = u
                binned_data[f'{b}{lum_ext}{self.upper_limit_key}'] = bin_ulim_bool
                binned_data[f'{b}{lum_ext}{self.Npoints_key}'] = n_points

                outlier_masks[lum_ext] = outlier_mask
                use_masks[lum_ext] = use_mask
                bin_ulim_bools[lum_ext] = bin_ulim_bool

            # -------  calculate the zeropoints per exposure ------- #
            # this might look wrong since we use the flux mask on the magnitudes but it s right
            # for each flux measurement we need the corresponding magnitude to get the zeropoint
            mags = lightcurve[f'{b}{self.mag_key_ext}']
            inst_fluxes = lightcurve[f'{b}{self.flux_key_ext}']
            pos_m = inst_fluxes > 0  # select only positive fluxes, i.e. detections
            zp_mask = pos_m & use_masks[self.flux_key_ext]

            # calculate zero points
            zps = np.zeros_like(inst_fluxes)
            zps[zp_mask] = mags[zp_mask] + 2.5 * np.log10(inst_fluxes[zp_mask])
            # find visits with no zeropoints
            n_valid_zps = np.bincount(visit_map, weights=zp_mask)
            at_least_one_valid_zp = n_valid_zps > 0
            # calculate the median zeropoint for each visit
            zps_median = np.zeros_like(n_valid_zps, dtype=float)
            zps_median[n_valid_zps > 0] = np.array([
                np.median(zps[(visit_map == i) & zp_mask])
                for i in np.unique(visit_map[at_least_one_valid_zp[visit_map]])
            ])
            # if there are only non-detections then fall back to default zeropoint
            zps_median[n_valid_zps == 0] = self.magnitude_zeropoints['Mag'][b]
            # if the visit only has upper limits then use the fall-back zeropoint
            zps_median[bin_ulim_bools[self.flux_key_ext]] = self.magnitude_zeropoints['Mag'][b]

            # ---------------   calculate flux density from instrument flux   ---------------- #
            # get the instrument flux [digital numbers], i.e. source count
            inst_fluxes_e = lightcurve[f'{b}{self.flux_key_ext}{self.error_key_ext}']

            # calculate the proportionality constant between flux density and source count
            mag_zp = self.magnitude_zeropoints['F_nu'][b].to('mJy').value
            flux_dens_const = mag_zp * 10 ** (-zps_median / 2.5)

            # calculate flux densities from instrument counts
            flux_densities = inst_fluxes * flux_dens_const[visit_map]
            flux_densities_e = inst_fluxes_e * flux_dens_const[visit_map]

            # bin flux densities
            mean_fd, u_fd, ul_fd, outlier_mask_fd, use_mask_fd, n_points_fd = self.calculate_epochs(
                flux_densities,
                flux_densities_e,
                visit_map, counts,
                remove_outliers=False,
                outlier_mask=
                outlier_masks[
                    self.flux_key_ext]
            )
            binned_data[f'{b}{self.mean_key}{self.flux_density_key_ext}'] = mean_fd
            binned_data[f'{b}{self.flux_density_key_ext}{self.rms_key}'] = u_fd
            binned_data[f'{b}{self.flux_density_key_ext}{self.upper_limit_key}'] = ul_fd
            binned_data[f'{b}{self.flux_density_key_ext}{self.Npoints_key}'] = n_points_fd

        return pd.DataFrame(binned_data)

    def calculate_metadata_single(self, lc):
        """
        Calculates some metadata, describing the variability of the lightcurves.

        - `max_dif`: maximum difference in magnitude between any two datapoints
        - `min_rms`: the minimum errorbar of all datapoints
        - `N_datapoints`: The number of datapoints
        - `max_deltat`: the maximum time difference between any two datapoints
        - `mean_weighted_ppb`: the weighted average brightness where the weights are the points per bin

        :param lc: the lightcurves
        :type lc: dict
        :return: the metadata
        :rtype: dict
        """

        metadata = dict()

        for band in self.bands:
            for lum_key in [self.mag_key_ext, self.flux_key_ext, self.flux_density_key_ext]:
                llumkey = f"{band}{self.mean_key}{lum_key}"
                errkey = f"{band}{lum_key}{self.rms_key}"
                ul_key = f'{band}{lum_key}{self.upper_limit_key}'
                ppb_key = f'{band}{lum_key}{self.Npoints_key}'

                difk = f"{band}_max_dif{lum_key}"
                rmsk = f"{band}_min_rms{lum_key}"
                Nk = f"{band}_N_datapoints{lum_key}"
                dtk = f"{band}_max_deltat{lum_key}"
                medk = f"{band}_median{lum_key}"
                chi2tmk = f"{band}_chi2_to_med{lum_key}"
                mean_weighted_ppb_key = f"{band}_mean_weighted_ppb{lum_key}"
                excess_variance_key = f"{band}_excess_variance_{lum_key}"
                excess_variance_err_key = f"{band}_excess_variance_err_{lum_key}"
                mck = f"{band}_coverage_of_median{lum_key}"

                ilc = lc[~np.array(lc[ul_key]).astype(bool)] if ul_key in lc else dict()
                metadata[Nk] = len(ilc)

                if len(ilc) > 0:
                    # check if ppb sum to more than zero
                    if sum(ilc[ppb_key]) > 0:
                        metadata[mean_weighted_ppb_key] = np.average(ilc[llumkey], weights=ilc[ppb_key])
                        metadata[excess_variance_key], metadata[excess_variance_err_key] = get_excess_variance(
                            np.array(ilc[llumkey]),
                            np.array(ilc[errkey]),
                            np.array(metadata[mean_weighted_ppb_key])
                        )

                    # if not we can not calculate the excess variance
                    else:
                        for k in [mean_weighted_ppb_key, excess_variance_key, excess_variance_err_key]:
                            metadata[k] = np.nan

                    imin = ilc[llumkey].min()
                    imax = ilc[llumkey].max()
                    imin_rms_ind = ilc[errkey].argmin()
                    imin_rms = ilc[errkey].iloc[imin_rms_ind]

                    imed = np.median(ilc[llumkey])
                    ichi2_to_med = sum(((ilc[llumkey] - imed) / ilc[errkey]) ** 2)
                    imc = np.sum(abs(ilc[llumkey] - imed) < ilc[errkey]) / len(ilc[llumkey])

                    metadata[difk] = imax - imin
                    metadata[rmsk] = imin_rms
                    metadata[medk] = imed
                    metadata[chi2tmk] = ichi2_to_med
                    metadata[mck] = imc

                    if len(ilc) == 1:
                        metadata[dtk] = 0
                    else:
                        mjds = np.array(ilc.mean_mjd).astype(float)
                        dt = mjds[1:] - mjds[:-1]
                        metadata[dtk] = max(dt)

                else:
                    for k in [
                        difk,
                        dtk,
                        mean_weighted_ppb_key,
                        excess_variance_key,
                        rmsk,
                        medk,
                        chi2tmk

                    ]:
                        metadata[k] = np.nan

        return metadata

    def plot_diagnostic_binning(
            self,
            service,
            ind,
            lum_key="mag",
            interactive=False,
            fn=None,
            save=True,
            which="panstarrs",
            arcsec=20
    ):
        """
        Show a skymap of the single detections and which bin they belong to next to the binned lightcurve

        :param service: service used to download data, either of 'tap' or 'gator'
        :type service: str
        :param ind: index of the object in the parent sample
        :type ind: str, int
        :param lum_key: the key of the brightness unit, either of `flux` (instrument flux in counts) or `mag`
        :type lum_key: str
        :param interactive: if function is used interactively, return mpl.Figure and mpl.axes if True
        :type interactive: bool
        :param fn: filename for saving
        :type fn: str
        :param save: saves figure if True
        :type save: bool
        :param which: survey to get the cutout from, either of 'sdss' or 'panstarrs'
        :type which: str
        :param arcsec: size of cutout
        :type arcsec: float
        :returns: Figure and axes if `interactive=True`
        :rtype: mpl.Figure, mpl.Axes
        """

        logger.info(f"making binning diagnostic plot")
        pos = self.parent_sample.df.loc[
            ind,
            [self.parent_sample.default_keymap["ra"], self.parent_sample.default_keymap["dec"]]
        ]
        chunk_number = self._get_chunk_number(parent_sample_index=ind)

        if service == "tap":
            unbinned_lcs = self.get_unbinned_lightcurves(chunk_number=chunk_number)
        else:
            unbinned_lcs = self._get_unbinned_lightcurves_gator(chunk_number=chunk_number)

        lightcurve = unbinned_lcs[unbinned_lcs[self._tap_orig_id_key] == ind]
        binned_lightcurve = self.bin_lightcurve(lightcurve)

        fig, axs = plt.subplots(nrows=2, gridspec_kw={"height_ratios": [3, 2]}, figsize=(5, 8))

        kwargs = {"plot_color_image": True} if which == "panstarrs" else dict()
        self.parent_sample.plot_cutout(ind=ind, ax=axs[0], which=which, arcsec=arcsec, **kwargs)
        self._plot_lc(lightcurve=binned_lightcurve, unbinned_lc=lightcurve, lum_key=lum_key, ax=axs[-1], save=False)
        axs[-1].set_ylabel("Apparent Vega Magnitude")
        axs[-1].grid(ls=":", alpha=0.5)

        visit_map = self.get_visit_map(lightcurve)

        markers = ['o', 'v', '^', '<', '>', '1', '2', '3', '4', '8', 's', 'p', '*', 'h', 'H', '+', 'x', 'D', 'd', '|',
                   '_', 'P', 'X']

        for visit in np.unique(visit_map):
            m = visit_map == visit
            datapoints = lightcurve[m]

            label = str(visit)
            marker = markers[visit]
            color = f"C{visit}"
            ra = (datapoints.ra - pos[self.parent_sample.default_keymap["ra"]]) * 3600
            dec = (datapoints.dec - pos[self.parent_sample.default_keymap["dec"]]) * 3600

            if ("sigra" in datapoints.columns) and ("sigdec" in datapoints.columns):
                has_sig = ~datapoints.sigra.isna() & ~datapoints.sigdec.isna()
                axs[0].errorbar(
                    ra[has_sig],
                    dec[has_sig],
                    xerr=datapoints.sigra[has_sig] / 3600,
                    yerr=datapoints.sigdec[has_sig] / 3600,
                    label=label,
                    marker=marker,
                    ls="",
                    color=color
                )
                axs[0].scatter(
                    datapoints.ra[~has_sig],
                    datapoints.dec[~has_sig],
                    marker=marker,
                    color=color
                )
            else:
                axs[0].scatter(ra, dec, label=label, marker=marker, color=color)

        title = axs[0].get_title()
        axs[0].set_title("")
        axs[0].legend(ncol=5, bbox_to_anchor=(0, 1, 1, 0), loc="lower left", mode="expand", title=title)
        axs[0].set_aspect(1, adjustable="box")
        fig.tight_layout()

        if save:
            if fn is None:
                fn = os.path.join(self.plots_dir, f"{ind}_binning_diag_{which}cutout.pdf")
            logger.debug(f"saving under {fn}")
            fig.savefig(fn)

        if interactive:
            return fig, axs
        else:
            plt.close()
